#include <errno.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <netinet/in.h>

#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>

#include "qproxy.h"
#include "funcs.h"
#include "io.h"

struct buffer_t *cl_in, *sv_in, *cl_out, *sv_out;
struct buffer_t *cl_pending, *sv_pending;
int clientfd, serverfd;
struct sockaddr_in client_l, client_r, server_l, server_r;
extern int verbose;
extern int cli_seqnum, cli_seqack, ser_seqnum, seq_seqack;

void qw_connect(char *host, int remote_port, int local_port)
{
	struct hostent *he;

	if (!(he = gethostbyname(host))) {
		fprintf(stderr, "host name lookup of %s failed\n", host);
		return;
	}
	memset(&client_l, 0, sizeof(struct sockaddr_in));

	client_l.sin_family = AF_INET;
	client_l.sin_port = htons(local_port);
	client_l.sin_addr.s_addr = htons(INADDR_ANY);

	memset(&client_r, 0, sizeof(struct sockaddr_in));

	memset(&server_l, 0, sizeof(struct sockaddr_in));

	server_l.sin_family = AF_INET;
	server_l.sin_addr.s_addr = htons(INADDR_ANY);

	memset(&server_r, 0, sizeof(struct sockaddr_in));

	server_r.sin_family = AF_INET;
	server_r.sin_port = htons(remote_port);
	memcpy(&server_r.sin_addr, he->h_addr, he->h_length);

	if ((clientfd = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP)) < 0) {
		perror("client socket()");
		return;
	}
	if (bind(clientfd, (struct sockaddr *)&client_l, sizeof(struct sockaddr_in)) < 0) {
		perror("client bind()");
		return;
	}
	if ((serverfd = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP)) < 0) {
		perror("server socket()");
		return;
	}
	if (bind(serverfd, (struct sockaddr *)&server_l, sizeof(struct sockaddr_in)) < 0) {
		perror("server bind()");
		return;
	}
	cl_in = malloc(sizeof(struct buffer_t));
	cl_out = malloc(sizeof(struct buffer_t));
	cl_pending = malloc(sizeof(struct buffer_t));
	sv_in = malloc(sizeof(struct buffer_t));
	sv_out = malloc(sizeof(struct buffer_t));
	sv_pending = malloc(sizeof(struct buffer_t));

	cl_in->bufmem = cl_in->buf = malloc(BUF_SIZE);
	cl_out->bufmem = cl_out->buf = malloc(BUF_SIZE);
	cl_pending->bufmem = cl_pending->buf = malloc(BUF_SIZE);
	sv_in->bufmem = sv_in->buf = malloc(BUF_SIZE);
	sv_out->bufmem = sv_out->buf = malloc(BUF_SIZE);
	sv_pending->bufmem = sv_pending->buf = malloc(BUF_SIZE);

	memset(cl_pending->bufmem, 0, BUF_SIZE);
	memset(sv_pending->bufmem, 0, BUF_SIZE);

	cl_pending->len = cl_pending->pos = 0;
	sv_pending->len = sv_pending->pos = 0;
}

void qw_disconnect(void)
{
	close(clientfd);
	close(serverfd);

	free(cl_in->bufmem);
	free(cl_out->bufmem);
	free(cl_pending->bufmem);
	free(sv_in->bufmem);
	free(sv_out->bufmem);
	free(sv_pending->bufmem);
	
	free(cl_in);
	free(cl_out);
	free(cl_pending);
	free(sv_in);
	free(sv_out);
	free(sv_pending);
}

void mainloop(void)
{
	struct timeval tv;
	fd_set fds;
	int sockaddr_len = sizeof(struct sockaddr_in);

	for (;;) {
		FD_ZERO(&fds);
		FD_SET(clientfd, &fds);
		FD_SET(serverfd, &fds);

		cl_in->buf = cl_in->bufmem;
		cl_out->buf = cl_out->bufmem;
		sv_in->buf = sv_in->bufmem;
		sv_out->buf = sv_out->bufmem;

		memset(cl_in->buf, 0, BUF_SIZE);
		memset(cl_out->buf, 0, BUF_SIZE);
		memset(sv_in->buf, 0, BUF_SIZE);
		memset(sv_out->buf, 0, BUF_SIZE);

		cl_in->len = cl_out->len = 0;
		sv_in->len = sv_out->len = 0;

		cl_in->pos = cl_out->pos = 0;
		sv_in->pos = sv_out->pos = 0;

		tv.tv_sec = 60;
		tv.tv_usec = 0;

		if (select(256, &fds, NULL, NULL, &tv) < 0 && errno != EINTR)
			perror("warning: select()");

		if (FD_ISSET(clientfd, &fds)) {
			if ((cl_in->len = recvfrom(clientfd, cl_in->bufmem, BUF_SIZE, 0, (struct sockaddr *)&client_r, &sockaddr_len)) < 0)
				perror("client->server recvfrom()");
			else {
				if (verbose) {
					fprintf(stderr, "client in: ");
					print_data(cl_in);
				}
			}
		}

		if (FD_ISSET(serverfd, &fds)) {
			if ((sv_in->len = recvfrom(serverfd, sv_in->bufmem, BUF_SIZE, 0, (struct sockaddr *)&server_r, &sockaddr_len)) < 0)
				perror("server->client recvfrom()");
			else {
				if (verbose) {
					fprintf(stderr, "server in: ");
					print_data(sv_in);
				}
			}
		}

		if (sv_in->len > 0)
			server_parse();
		if (cl_in->len > 0)
			client_parse();

		check_timers();

		if (cl_in->len) {
			if (cl_pending->len && cli_seqnum != 0xffffffff) {
				memcpy(cl_out->bufmem + cl_out->len, cl_pending->bufmem, cl_pending->len);
				cl_out->len += cl_pending->len;
				if (verbose) {
					fprintf(stderr, "client pending: ");
					print_data(cl_pending);
				}

			}
			if (sendto(serverfd, cl_out->bufmem, cl_out->len, 0, (struct sockaddr *)&server_r, sockaddr_len) < 0)
				perror("client->server sendto()");
			else {
				if (verbose) {
					fprintf(stderr, "client out: ");
					print_data(cl_out);
				}
			}

			if (cl_pending->len && cli_seqnum != 0xffffffff) {
				memset(cl_pending->bufmem, 0, BUF_SIZE);
				cl_pending->buf = cl_pending->bufmem;
				cl_pending->len = cl_pending->pos = 0;
			}
		} else if (cl_out->len) {
			memcpy(cl_pending->bufmem + cl_pending->len, cl_out->bufmem, cl_out->len);
			cl_pending->len += cl_out->len;
		}

		if (sv_in->len) {
			if (sv_pending->len && ser_seqnum != 0xffffffff) {
				if (verbose) {
					fprintf(stderr, "server pending: ");
					print_data(sv_pending);
				}
				memcpy(sv_out->bufmem + sv_out->len, sv_pending->bufmem, sv_pending->len);
				sv_out->len += sv_pending->len;
			}
			if (sendto(clientfd, sv_out->bufmem, sv_out->len, 0, (struct sockaddr *)&client_r, sockaddr_len) < 0)
				perror("server->client sendto()");
			else {
				if (verbose) {
					fprintf(stderr, "server out: ");
					print_data(sv_out);
				}
			}
			if (sv_pending->len && ser_seqnum != 0xffffffff) {
				memset(sv_pending->bufmem, 0, BUF_SIZE);
				sv_pending->buf = sv_pending->bufmem;
				sv_pending->len = sv_pending->pos = 0;
			}
		} else if (sv_out->len) {
			memcpy(sv_pending->bufmem + sv_pending->len, sv_out->bufmem, sv_out->len);
			sv_pending->len += sv_out->len;
		}
		
	}
}


unsigned char get_byte(struct buffer_t *data)
{
	unsigned char x;

	x = *data->buf;
	data->buf++;
	data->pos++;
	return x;
}

void put_byte(struct buffer_t *data, unsigned char byte)
{
	*data->buf = byte;
	data->buf++;
	data->len++;
	data->pos++;
}

char *get_string(struct buffer_t *data)
{
	static char buf[2048];
	int x = 0;

	while (*data->buf) {
		buf[x] = get_byte(data);
		x++;
		if (x > 2047)
			break;
	}
	buf[x] = get_byte(data);
	return buf;
}

void put_string(struct buffer_t *data, char *string)
{
	if (strlen(string) > 2047)
		string[2048] = 0;

	while (*string) {
		put_byte(data, *string);
		string++;
	}
	put_byte(data, 0);
}

long get_long(struct buffer_t *data)
{
	long x;

	x = get_byte(data);
	x |= get_byte(data) << 8;
	x |= get_byte(data) << 16;
	x |= get_byte(data) << 24;

	return x;
}

void put_long(struct buffer_t *data, long x)
{
	put_byte(data, x & 0xff);
	put_byte(data, (x >> 8) & 0xff);
	put_byte(data, (x >> 16) & 0xff);
	put_byte(data, (x >> 24) & 0xff);
}

short get_short(struct buffer_t *data)
{
	short x;

	x = get_byte(data);
	x |= get_byte(data) << 8;

	return x;
}

void put_short(struct buffer_t *data, short x)
{
	put_byte(data, x & 0xff);
	put_byte(data, (x >> 8) & 0xff);
}

float get_coord(struct buffer_t *data)
{
	return (float)(get_short(data) * 0.125);
}

void put_coord(struct buffer_t *data, float x)
{
	put_short(data, x / 0.125);
}

float get_angle(struct buffer_t *data)
{
	return (float)((get_byte(data) / 256.0) * 360.0);
}

void put_angle(struct buffer_t *data, float x)
{
	put_byte(data, ((x / 360.0) * 256.0));
}

float get_angle16(struct buffer_t *data)
{
	return (float)((get_short(data) / 65536.0) * 360.0);
}

void put_angle16(struct buffer_t *data, float x)
{
	put_short(data, ((x / 360.0) * 65536.0));
}

float get_float(struct buffer_t *data)
{
	char buf[4];
	float x;
	int pos;

	for (pos = 0; pos < 4; pos++)
		buf[pos] = get_byte(data);

	memcpy(&x, buf, sizeof(x));
	return x;
}

void put_float(struct buffer_t *data, float x)
{
	char buf[4];
	int pos;

	memcpy(buf, &x, sizeof(x));

	for (pos = 0; pos < 4; pos++)
		put_byte(data, buf[pos]);
}

void print_data(struct buffer_t *data)
{
#ifdef DEBUG
	int x;
	
	if (verbose > 1) {
		write(2, data->bufmem, data->len);
		write(2, "\n", 1);
	}
	if (verbose) {
		for (x = 0; x < data->len; x++) 
			fprintf(stderr, "%02x ", data->bufmem[x]);
		fprintf(stderr, "\n");
	}
#endif
}
